use rusqlite::{Connection, Row, ToSql};
use serde::{Deserialize, Serialize};

use super::error::DataStoreError;

#[derive(Debug, Copy, Clone, PartialEq, Deserialize, Serialize)]
pub enum SortOrder {
    ASC,
    DESC,
}

pub struct Pagination<'a> {
    pub sort_col: &'a str,
    pub limit: u64,
    pub cursor: Option<&'a str>,
    pub sort_order: SortOrder,
}

impl<'a> Pagination<'a> {
    pub fn new(
        sort_col: &'a str,
        cursor: Option<&'a str>,
        limit: u64,
        sort_order: SortOrder,
    ) -> Self {
        Self {
            sort_col,
            limit,
            cursor,
            sort_order,
        }
    }

    pub fn paginate<T, F>(
        &self,
        conn: &Connection,
        statement: &str,
        params: &[(&str, &dyn ToSql)],
        processor: F,
    ) -> Result<Vec<T>, DataStoreError>
    where
        F: FnMut(&Row<'_>) -> Result<T, rusqlite::Error>,
    {
        let cursor = self.cursor.unwrap_or("");
        let comparison = if self.sort_order == SortOrder::ASC {
            ">"
        } else {
            "<"
        };
        let where_sql = if statement.ends_with("AND") {
            ""
        } else {
            "WHERE"
        };

        let sort_col = self.sort_col;
        // otherwise paginated statements with join will fail
        let sort_col_wrapped = if let Some(index) = sort_col.find('.') {
            let right_part = &sort_col[index..];
            "t".to_owned() + right_part
        } else {
            sort_col.to_owned()
        };
        let sort_order = &self.sort_order;
        // DANGER: vulnerable to SQL injection if statement or sort_col variable is influenced by user input
        let mut stmt = conn.prepare(&format!(
            "SELECT * FROM ({statement} {where_sql} {sort_col} {comparison} (:cursor) ORDER BY {sort_col} {sort_order:?} LIMIT (:limit)) AS t ORDER BY {sort_col_wrapped} ASC"
        ))?;
        let params = [
            &[
                (":cursor", &cursor as &dyn ToSql),
                (":limit", &self.limit as &dyn ToSql),
            ],
            params,
        ]
        .concat();
        let iter = stmt.query_map(params.as_slice(), processor)?;
        Ok(iter.filter_map(Result::ok).collect())
    }
}
